use crate::scoring::ScoringMode;

pub const FSCORING_CMD: &str = "fscoring";
pub const DEFAULT_OUT: &str = "fscoring.csv.gz";
pub const DEFAULT_SCORING_MODE: ScoringMode = ScoringMode::Atac;
pub const START_SHIFT: i8 = 4;
pub const END_SHIFT: i8 = 5;
