from typing import List

def write_tokens_to_gtok(filename: str, tokens: List[str]):
    """
    Write tokens to a GTOK file.

    :param filename: The filename of the GTOK file.
    :param tokens: The tokens to write.
    """
    pass

def read_tokens_from_gtok(filename: str) -> List[str]:
    """
    Read tokens from a GTOK file.

    :param filename: The filename of the GTOK file.
    """
    pass
